! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: rgsangles
!
! Authors:  Peter Videler - SOC ESTEC, C. Erd, XMM-SOC
!          
! Description: Correct RFC events for spacecraft poiinting
! 
! Comments: 
! $Id: rgsangles_mod.f90,v 1.57 2013/11/05 13:35:50 aibarra Exp $
!- 
module rgsangles_mod

  use param  
  use saslib
  use sasdal
  use constants
  use odfNames
  use errorHandling
  use rgslib
  use pointing_mod       ! rgslib/lib/pointing_mod.f90
  use euler321_mod       ! rgslib/lib/euler321_mod.f90
  use CAL
  use oal
  use barycen      ! barycen/lib/heliocorr_mod

  implicit none 

  private

  ! ......................................................
  public rgsangles
  ! ......................................................

  ! Data structure for event table.
  ! Only need to define columns needed 

  type eventArray_t
     real (single),     pointer :: detX(:), detY(:), detZ(:)
     real (double),     pointer :: time(:)
     real (single),     pointer :: mlambda(:)
     real (single),     pointer :: beta(:), betaraw(:)
     real (single),     pointer :: xdsp(:), xdspraw(:)
     real (single),     pointer :: pointingRA(:), pointingDEC(:), pointingPOS(:) 
     integer(int8),     pointer :: ccdId(:)
  endType eventArray_t

  ! globals
   type (eventArray_t)           :: e            ! event table
   type (TableT)                 :: tbl          ! events table

contains 

  !---------------------------------------------------------------------
  subroutine rgsangles
    !---------------------------------------------------------------------
    use vector_mod

    type(DataSetT)                :: set, srcset
    type(TableT)                  :: frames, srctbl 
    type(ColumnT)                 :: beta_col, xdsp_col

    type(SkyPositionType)         :: srcPos
    type(SpacecraftAttitudeType)  :: att0,att       ! typedef from oal
    character(len=*), parameter   :: fn='rgsangles: '
    integer                       :: ccdId, index, nbins, i
    real (single)                 :: binRef, binWidth

    character(len=10)   :: tblname
    logical, allocatable, dimension(:) :: hasatt
    integer, allocatable, dimension(:) :: attindex
  
    set = dataSet (stringParameter("mergedset") , MODIFY)

    call CAL_setState(instrument=instId(stringAttribute(set,"INSTRUME")))
    call RGSLIB_refreshCache()

    tbl = table(set, RGS_EVENTS_TABLE_NAME )
    write(errStr,*) fn//'Processing table with ',numberOfRows(tbl),' events.'
    call message(appMsg, Verbose)

    ! time
    e%time =>  real64Data(Column(tbl, TIME_COLUMN_NAME , read))

    ! dispersion direction
    e%betaraw   => real32Data(Column(tbl,'BETA', read))
    beta_col = addColumnOverwrite(tbl,'BETA_CORR' , real32, 'rad')
    e%beta      => real32Data(beta_col)

    ! cross dispersion optional
    if (hasColumn(tbl,'XDSP' )) then
       e%xdspRaw   => real32Data(Column(tbl,'XDSP', read))
       xdsp_col = addColumnOverwrite(tbl,'XDSP_CORR', real32, 'rad')
       e%xdsp   => real32Data(xdsp_col)
    endif

    ! optional colums

    if (booleanParameter("withmlambdacolumn")) then
       e%mlambda    => real32Data(addColumnOverwrite(tbl, 'M_LAMBDA',     real32, 'Angstrom'))
    endif

    if (booleanParameter("withpointingcolumn")) then
       e%pointingRA  => real32Data(addColumnOverwrite(tbl,'SC_POINTING_RA' , real32, 'rad'))
       e%pointingDEC => real32Data(addColumnOverwrite(tbl,'SC_POINTING_DEC', real32, 'rad'))
       e%pointingPOS => real32Data(addColumnOverwrite(tbl,'SC_POINTING_POS', real32, 'rad'))
    endif

    ! ------------------------
    ! get reference attitude from table header
    ! ------------------------

     srcset = dataSet(stringParameter("srclist"), READ)
     srctbl = table(srcset, "SRCLIST", READ)
     call retrieve(srcset, att0)   ! retrieve reference attitude

     write(errstr,*) 'Reference Attitude used Ra Dec Apos (degrees) =', degrees(att0)
     call message(appMsg, Sparse)

                                !store reference attitude in events table header
     call setAttribute(tbl,"RA_PNT"  , att0%ra  /deg2rad, "deg","RA of instrument pointing")
     call setAttribute(tbl,"DEC_PNT" , att0%dec /deg2rad, "deg","DEC of instrument pointing")
     call setAttribute(tbl,"PA_PNT", att0%apos/deg2rad, "deg","Position Angle of instrument pointing")
     call setAttribute(tbl,"RADECSYS"  , "FK5", "stellar reference frame used")
     call setAttribute(tbl,"EQUINOX", 2000.0, "", "year of epoch")
                                !and in primary header
     call setAttribute(set,"RA_PNT"  , att0%ra  /deg2rad, "deg","RA of instrument pointing")
     call setAttribute(set,"DEC_PNT" , att0%dec /deg2rad, "deg","DEC of instrument pointing")
     call setAttribute(set,"PA_PNT", att0%apos/deg2rad, "deg","Position Angle of instrument pointing")
     call setAttribute(set,"RADECSYS"  , "FK5", "stellar reference frame used")
     call setAttribute(set,"EQUINOX", 2000.0, "", "year of epoch")
!     call store(tbl, att0)         ! 

    ! ------------------------
    ! select Source Position
    ! ------------------------

    srcPos%ra=0 ; srcPos%dec=0
    index =  Int32Attribute(srctbl,"PRIMESRC")
    srcPos = RGSLIB_getsourcePos(srctbl, index=index)
    call release (srcset)

   ! store source position used by rgsangles in events table header
    call store(tbl, srcPos)

    ! rgsexposure 1.12 wants attribute SOURCEID in events table.
    call setAttribute(tbl,"SourceId"  ,index, "","Source ID number in SOURCES table")
 
    write(errstr,*) 'Sourceposition used Ra Dec (degrees) =', degrees(srcPos)
    call message(appMsg, Sparse)

    ! ------------------------
    ! process exposure table (what used to be frames table)
    ! ------------------------

    do i=1,9                    !loop over EXPOSU0n tables (one per CCD)
       write (tblname,"('EXPOSU',i2.2)") i
       if ( hasBlock(set, tblname) ) then
          frames = table(set, tblname)
          call CAL_getState(ccdChipId=ccdId) !just to be sure, we setup the CAL here
          call processFramesTable(frames=frames, referenceAttitude=att0)
          call release (frames)
       endif
    end do

    ! ------------------------
    ! process events table
    ! ------------------------

    allocate( hasatt(0:size(e%beta)-1) )
                                !need to loop to catch non-existent attitude
    do i=0,size(e%beta)-1
       if ( OAL_getAttitude(e%time(i),att) ) then
          hasatt(i) = .true.
       else
          hasatt(i) = .false.
          call setToNull( beta_col, i)
          if ( associated(e%xdsp) ) call setToNull( xdsp_col, i)

	  call warning('No Attitude available',&
	  &'No attitude available for a certain row of event list')
!          write(errstr,*) 'No Attitude available for row',i,'of event list.'
!          call message(appMsg, Sparse)
       end if
    end do

    allocate( attindex(count(hasatt)) ) !collect all indices with valid attitude
    attindex = pack( (/ (i,i=0,size(hasatt)-1) /), hasatt)

    call processEvents(sourcePos=srcPos, referenceAttitude=att0, attindex=attindex)

    deallocate( attindex )

    ! ---------------
    ! binning of beta
    ! ---------------
    !poperties of channels
    nbins = intParameter("nbetabins")
    select case (stringParameter("betabinning")) 
    case ('Range')
       binWidth = (realParameter("betamax") - realParameter("betamin")) / nbins
       binref = realParameter("betamin") + 0.5*binWidth
    case ('binSize')
       binRef   = realParameter("betabinref")
       binWidth = realParameter("betabinwidth")
    case default
       call error (fn//'case not found. betabinning') 
    end select
    

    call angles2channels( nbins, binref, binWidth, 'BETA', e%beta, hasatt )

    !AI: Wavelength scale. Start
    ! Default lambda values
    ! mLambdamax = 40 Ao
    ! mLambdamin = 4 Ao    
    ! binWidth = 0.010 !Ao =  10 mAo
    binWidth = realParameter("lambdabinwidth") !Ao =  10 mAo
    !binref = 4.0 + 0.5*binWidth
    binref = realParameter("lambdabinref")
    nbins = intParameter("nlambdabins")
   
    if (booleanParameter("withmlambdacolumn")) then
    	call angles2channels( nbins, binref, binWidth, 'MLAMBDA', e%mlambda, hasatt)
    endif
 	!AI: Wavelength scale. Start

    ! ---------------
    ! binning of xdsp
    ! ---------------

    if (associated(e%xdsp)) then
                                !poperties of channels
       nbins = intParameter("nxdispbins")
       select case (stringParameter("xdispbinning")) 
       case ('Range')
          binWidth = (realParameter("xdispmax") - realParameter("xdispmin")) / nbins
          binref = realParameter("xdispmin") + 0.5*binWidth
       case ('binSize')
          binRef   = realParameter("xdispbinref")
          binWidth = realParameter("xdispbinwidth")
       case default
          call error (fn//'case not found. xdispbinning') 
       end select

       call angles2channels( nbins, binref, binWidth, 'XDSP', e%xdsp, hasatt )
    end if

    deallocate( hasatt )

    ! ------------------------
    ! Copy Attributes to make other packages happy. 
    ! ------------------------

!v2.0    if (.not. hasAttribute(tbl,"TELAPSE")) then
!v2.0       telapse = ToTimeTag(exposureInfo%actualEndTime) & 
!v2.0            - ToTimeTag(exposureInfo%actualStartTime)
!v2.0       call setAttribute(tbl, "TELAPSE", telapse,"sec", "Time eleapsed between start and end of observation")
!v2.0       ! call copyAttribute(attribute(tbl,"TELAPSE"), attributable(set))
!v2.0
!v2.0       call warning('telapse not found','Keyword TELAPSE not found in EVENTS table. Adding it.')
!v2.0    endif

    ! ------------------------
    !  finish
    ! ------------------------

    call release (tbl)
    call release (set)
  end subroutine rgsangles

  !---------------------------------------------------------------------
  subroutine processEvents(sourcePos, referenceAttitude, attindex)
    !---------------------------------------------------------------------

    type(SkyPositionType),        intent(in)  :: sourcePos        ! source Position
    type(SpacecraftAttitudeType), intent(in)  :: referenceAttitude   ! type def from OAL
    integer, intent(in), dimension(:) :: attindex !event index with good attitude

    real (single), dimension(size(e%time),2) :: d_angles
    type(SpacecraftAttitudeType)           :: att              ! type def from OAL

    real (single)                      :: FoverL, dummy(2), Qfactor
    real (single)                      :: alpha0, phi0dsp, phi0xdsp
    integer                            :: i
    integer(int32) :: instr
    real(kind=double)  :: SAA, vFactor, cteA, slopeB

    FoverL = RGSLIB_getFoverL()
    alpha0 = RGSLIB_getAlpha()

    !---------------
    ! off axis angle
    !--------------

    dummy = offAxisAngles(referenceAttitude, sourcePos) 
    phi0dsp  = dummy(1)
    phi0xdsp = dummy(2)

    ! write(*,*) 'reference attitude (rad)= ',referenceattitude   
    ! write(*,*) 'sourcePos (rad)= ',sourcePos   
    ! write(*,*) 'phi0=',   phi0dsp, phi0xdsp 
    ! write(*,*) 'alpha0=', alpha0 
    ! write(*,*) 'FoverL=', FoverL 

    ! See minutes of RGS analysis meeting #2. Email CE 02jul99

!    Qfactor = FoverL * sin(alpha0 + phi0dsp*FoverL)  ! CE 7-6-2000: phi0dsp already contains FoverL
    Qfactor = FoverL * sin(alpha0 + phi0dsp) 
    call setAttribute(tbl, "ASPFACT1"  ,Qfactor, "", &
         "Source dependent scaling factor used in Aspect Correction")

    !AI: Wavelength scale. Start
    ! DELETE NOT NEEDED
	call setAttribute(tbl, "OFFAXIS"  ,phi0dsp, "deg", &
         "Off Axis Angle")
        
    !AI: Wavelength scale. End 

    ! rest for non empty events tables only

    if (numberOfRows(tbl) > 0 ) then 

       ! ----------------------
       ! correction angles for spacecraft pointing
       ! ----------------------

       d_angles(attindex+1,:) = gratingAspectAngles(time=e%time(attindex), referenceAttitude=referenceAttitude)

       ! write(*,*) 'gratingAspectAngles  disp=', d_angles(:,1)
       ! write(*,*) 'gratingAspectAngles xdisp=', d_angles(:,2)

       !---------------
       ! corrected beta (dispersion direction)
       !---------------

       ! See minutes of RGS analysis meeting #2. Email CE 02jul99

       e%beta = e%betaraw + (Qfactor * d_angles(:,1) / sin(e%betaraw))

       !-----------------
       ! cross dispersion    
       !-----------------

       !write(*,*) 'phi0xdsp=', phi0xdsp
       if (associated(e%xdsp)) e%xdsp =  e%xdspRaw + d_angles(:,2) 

       ! ----------------------
       ! spacecraft pointing (optional, debug purpose)
       ! ----------------------

       if (associated(e%pointingRA)) then  
!          do i= 0, size(e%time) -1           ! note: dal starts at 0
          do i= 1, size(attindex)
             att = instrumentPointing(e%time(attindex(i)))
             e%pointingRA (attindex(i)) = att%ra  
             e%pointingDEC(attindex(i)) = att%dec
             e%pointingPOS(attindex(i)) = att%apos
          enddo
       endif

       ! ----------------------
       ! m*lambda (optional)
       ! ----------------------

       if (associated(e%mlambda)) then  

          e%mlambda  = beta2lambda(beta=e%beta, order=-1, offaxis=phi0dsp) *10. !convert to A

          !----------
          ! Call to barycen to calculate the Heliocentric corrections
          ! both Solar Aspect Angle and Heliocentric correction.
          !-----------
          instr = instId(stringAttribute(tbl,"INSTRUME"))
          cteA = 0.
          slopeB = 0.
          SAA = 0.
          vFactor =0.
          !-----------------
          ! Solar Aspect Angle correction
          !-----------------
          if (booleanParameter("withsunanglecorr")) then
              write(errstr,*) 'Calculating Solar Aspect Angle Correction'
              call message(appMsg, Sparse)
              if (.not.hasAttribute(tbl,"SUNANGCO")) then
                  if (HELIO_getSaaCorr(referenceAttitude%ra, referenceAttitude%dec,instr, SAA, cteA,slopeB)) then
                    SAA = Pi - SAA
                    write(errstr,*)'Solar angle value:  ',SAA, ' rad ', SAA/deg2rad, ' deg '
                    call message(appMsg, Sparse)
                    !write(*,*) 'Correction ',deg2rad ,SAA/deg2rad, ((cteA + slopeB*(SAA/deg2rad - 90.))/1000.)
                    e%mlambda = e%mlambda - ((cteA + slopeB*(SAA/deg2rad - 90.))/1000.)
                    call setAttribute(tbl,"SUNANGCO"  , "yes","Sun Angle correction applied")
                    call setAttribute(tbl,"SUNANGLE"  , SAA,"rad","Sun Angle")
                    call setAttribute(tbl,"CTEA"  , cteA," ","Sun angle correction cte")
                    call setAttribute(tbl,"SLOPEB"  , slopeB," ","Sun angle correction slope")
                  else
                    write(errstr,*)'Error calculating Solar Angle Correction'
                    call error("SAA_Error")
                  endif
                else
                  write(errstr,*)'Solar Angle correction already apply. Skipped '
                  call message(appMsg, Sparse)
              endif
          endif

          !------------------
          ! Helio centric velocity correction
          !
          !------------------
          if (booleanParameter("withheliocentriccorr")) then
            write(errstr,*) 'Calculating Heliocentric velocity Correction'
            call message(appMsg, Sparse)
            if (.not.hasAttribute(tbl,"HELIOCOR")) then
                if (HELIO_getVelCorr(referenceAttitude%ra, referenceAttitude%dec,vFactor)) then
                    write(errstr,*)'Heliocentric velocity correction factor ',vFactor
                    call message(appMsg, Sparse)
                    e%mlambda = e%mlambda*(1+ vFactor)
                    call setAttribute(tbl,"HELIOFAC", vFactor , "","Heliocentric velocity factor applied")

                else
                    write(errstr,*)'Error calculating Heliocentric velocity Correction'
                    call error("Helio_Error")
                endif
            else
                write(errstr,*)'Heliocentric velocity correction already apply. Skipped '
                call message(appMsg, Sparse)
            endif
          endif
       endif

    endif

  end subroutine processEvents

  !---------------------------------------------------------------------
  subroutine processFramesTable(frames, referenceAttitude)
    !---------------------------------------------------------------------
    type(TableT)                              :: frames
    type(ColumnT)                             :: aspDsp_col, aspXDsp_col
    type(SpacecraftAttitudeType), intent(in)  :: referenceattitude  ! type def from OAL

    type(SpacecraftAttitudeType) :: att
    real (single), pointer, dimension(:,:) :: dummy
    real (double),  pointer :: time(:)
    real (single),  pointer :: aspColDsp(:),aspColXDsp(:)
    integer :: i

    time =>  real64Data(Column(frames, TIME_COLUMN_NAME , read))

    allocate (dummy(0:size(time)-1, 2))

    aspDsp_col  = addColumnOverwrite(frames,'ASPCDSP' , real32, '')
    aspXDSP_col = addColumnOverwrite(frames,'ASPCXDSP', real32, '')
    aspColDsp  => real32Data(aspDsp_col)
    aspColXDsp => real32Data(aspXDsp_col)

    dummy = gratingAspectAngles(time=time, referenceAttitude=referenceAttitude)

    ! write(*,*) 'disp', dummy(:,1) 
    ! write(*,*) 'xdisp', dummy(:,2) 

    aspColDsp  = dummy(:,1)
    aspColXdsp = dummy(:,2)

    do i=0,size(time)-1
       if ( .not. OAL_getAttitude(time(i),att) ) then
          call setToNull( aspDsp_col , i )
          call setToNull( aspXDsp_col, i )
       end if
    end do

    deallocate(dummy)

    call release( aspDsp_col )
    call release( aspXDsp_col )
    call release( frames )
  end subroutine processFramesTable


  !---------------------------------------------------------------------
  subroutine angles2channels ( nbins, binRef, binWidth, type, angl, hasatt )
    !---------------------------------------------------------------------
                                !make binning into {BETA,XDSP,MLAMBDA}_CHANNEL
    integer, intent(in)          :: nbins
    real(single), intent(in)    :: binRef, binWidth
    character(len=*), intent(in) :: type
    real(single), intent(in), dimension(:) :: angl !angular data to be binned
    logical, intent(in), dimension(:) :: hasatt
    

    type(ColumnT) :: chan_col
    integer :: i, chan1, minchan, maxchan
    integer(int16), pointer :: chan(:)


    chan_col = addColumnOverwrite(tbl,type//'_CHANNEL', integer16, "")
    chan  => int16Data( chan_col )
    call setNullValue( chan_col, -32768) !smallest 16bit integer

    minchan = 2**15-1
    maxchan = -2**15

    do i=1,size(angl)
       if ( hasatt(i) ) then    !require valid attitude
!       e%beta_chan = asInt16(floor(0.5 + (e%beta - binRef) / betabinWidth))
          chan1 = nint(1. + (angl(i) - binRef) / binWidth)

          if ( chan1 < 1 .or. chan1 > nbins ) then
	     call warning("outside range","_CORR is outside the range specified by the binning parameters")	    
!             write(errstr,*) type//'_CORR is outside the range specified by the binning parameters, &
!                  &at row',i,',',type//'_CORR=',angl(i),type//'_CHANNEL=',chan1
!             call message(appMsg, Silent)
          end if

!
! gfortran does not handle NaN in the same way as NAGFor. Therefore, we must
! must check if angl is a valid number instead of chechking an overflow in the results. 
!
!          if ( chan1 < -2**15 .or. chan1 > 2**15-1 ) then !check int16 range  <-- NOT LONGER VALID IN

          if ( angl(i) .ne. angl(i) ) then !check angl is a valid number
	     call warning("binning ", "_CORR cannot be binned into channels")
!              write(errstr,*) type//'_CORR cannot be binned into channels at row',i,',',type//'_CORR=',angl(i)
!             call message(appMsg, Silent)
             call setToNull( chan_col, i-1 )
          else
             chan(i-1) = chan1
             minchan = min( minchan, chan1 )
             maxchan = max( maxchan, chan1 )           
          end if

       else                     !no attitude available
          call setToNull( chan_col, i-1 )
       end if
    end do


    call setAttribute(chan_col, "TDMIN", minchan, "", "minimum data value present")
    call setAttribute(chan_col, "TDMAX", maxchan, "", "maximum data value present")

    call setAttribute(chan_col, "TLMIN", 1,     "", "minimum legal channel number")
    call setAttribute(chan_col, "TLMAX", nbins, "", "maximum legal channel number")

    if ( type == 'BETA' ) then
       call setAttribute(chan_col, "TCTYP", "Dispersion Channel", "Name of axis")
    else if (type == 'XDSP' ) then
       call setAttribute(chan_col, "TCTYP", "Cross-dispersion Channel", "Name of axis")
    end if

    call setAttribute(chan_col, "TCUNI", "", "unit of channel (unitless)")
    call setAttribute(chan_col, "TCRPX", 1, "chan", "reference channel number")
    call setAttribute(chan_col, "TCRVL", binRef, "rad", "value at center of reference channel")
    call setAttribute(chan_col, "TCDLT", binWidth, "rad", "width of one channel")

    call release (chan_col)       
  end subroutine angles2channels
end module rgsangles_mod
